// Michael J. Wilson
// Music 256b Winter 2011
// Homework #3

#ifndef __SAXMAX_H__
#define __SAXMAX_H__

#include "mo_gfx.h"
#include "Saxofony.h"

const int SAMPLE_RATE = 44100;
const int FRAMESIZE = 512;
const int NUM_CHANNELS = 2;

namespace SaxMax
{
	// Setup and teardown
	void set_dims(GLint width, GLint height);
	bool init();
	void close();

	// Render one frame
	void render();

	// Reset control parameters to initial values.
	void reset_control_parameters();
	
	// Get / set STK parameter on instrument (see ControlParameter.h).  Value from 0.0-128.0
	void set_control(int parameter, float value);
	float get_control(int parameter);
	
	// Slew control values towards targets
	void slew_values();

	// Translate touch location to tonehole
	int get_tonehole_for_location(float x, float y);
	
	// Convert midi note to frequency
	float midi_to_frequency(int note);

	// Drawing helper functions
	void draw_waveform();
	void draw_toneholes();
	
	// Get/set reverb T60 (must be greater than 0)
	void set_reverb(float value);
	float get_reverb();
	
	// Calculate frequency based on currently engaged toneholes
	void calculate_frequency();
	
	// Return the instrument
	stk::Saxofony* get_instrument();
};
#endif