// Michael J. Wilson
// Music 256b Winter 2011
// Homework #3

#include "SaxMax.h"
#include "Saxofony.h"
#include "OnePole.h"
#include "mo_audio.h"
#include "mo_gfx.h"
#include "mo_touch.h"
#include "mo_accel.h"
#include "mo_fun.h"
#include "ControlParameter.h"
#include <vector>
#include "NRev.h"

#include <iostream>

// Screen dimensions
GLint g_gfxWidth = 0;
GLint g_gfxHeight = 0;

// Texture buffer
GLuint texture[1];

const int NUM_TONEHOLES = 5;

// For drawing waveform
const int NUM_VERTICES = 512;
float m_vertices[NUM_VERTICES];
GLfloat m_line_scratch[NUM_VERTICES*2];
int m_vertices_index = 0;

// STK objects
stk::Saxofony* instrument;
stk::OnePole* mic_filter;
stk::OnePole* mic_filter2;
stk::NRev* rev;

// Parameters
std::vector<ControlParameter*> control_parameters;
float reverb_t60 = 1.0;
float current_frequency = 440.0;

// Check if toneholes are down:
bool tonehole_status [NUM_TONEHOLES] = {false, false, false, false, false};

void accelCallback( double x, double y, double z, void * data )
{
//	Control Change Numbers: 
//	- Reed Stiffness = 2
//	- Reed Aperture = 26
//	- Noise Gain = 4
//	- Blow Position = 11
//	- Vibrato Frequency = 29
//	- Vibrato Gain = 1
//	- Breath Pressure = 128
	// These control parameters range from 0.0 to 128.0.
	
	float val = x;
	if (val > 1.0)
		val = 1.0;
	if(val < -1.0)
		val = -1.0;
	val += 1.0;
	val /= 2.0;
	val *= 128.0;
	
	control_parameters[BLOW_POSITION]->target = val;
}

int SaxMax::get_tonehole_for_location(float x, float y)
{
	int result = 0;
	if (x < g_gfxHeight/3.0)
	{
		if (y > g_gfxWidth/2.0)
		{
			result = 1;
		}
		else
		{
			result = 3;
		}
	}
	else if (x < g_gfxHeight*2.0/3.0)
	{
		if (y > g_gfxWidth/2.0)
		{
			result = 2;
		}
		else
		{
			result = 4;
		}				
	}
	else
	{
		result = 0;
	}
	return result;
}

void touchCallback( NSSet * touches, UIView * view, const std::vector<MoTouchTrack> & touchPts, void * data)
{
	// Clear the toneholes
	int i;
	for (i=0; i < NUM_TONEHOLES; ++i)
	{
		tonehole_status[i] = false;
	}
	
	// Track active toneholes
	for(i=0; i < touchPts.size(); ++i)
	{
		UITouch* touch = touchPts[i].touch;
		if( touch.phase == UITouchPhaseBegan ||
		    touch.phase == UITouchPhaseMoved ||
		    touch.phase == UITouchPhaseStationary)
		{
			CGPoint location = [touch locationInView:nil];
			double temp = location.x;
			location.x = location.y;
			location.y = temp;
			tonehole_status[SaxMax::get_tonehole_for_location(location.x, location.y)] = true;
		}
	}

	SaxMax::calculate_frequency();
}

void audio_callback(Float32* buffer, UInt32 numFrames, void* userData)
{
	static Float32 next_sample = 0.0;
	static Float32 next_breath_value = 0.0;
	static Float32 step_value = 0.0;
    for( int i = 0; i < numFrames; i++ )
    {
		// Get mic input as envelope for breath pressure.
		mic_filter->tick(abs(buffer[2*i]));
		//mic_filter2->tick(step_value);
		
		// Set next sample for the instrument
		next_sample = 0.0;
		next_sample += instrument->tick();
		
		// Reverb and stereo
        buffer[2*i] = rev->tick(next_sample, 0);
        buffer[2*i+1] = rev->tick(next_sample, 1);

		// Fill buffer
		m_vertices[m_vertices_index] = next_sample;
		m_vertices_index++;
		m_vertices_index %= NUM_VERTICES;
	}
	step_value = mic_filter->lastOut();
	next_breath_value = mic_filter2->tick(mic_filter->lastOut());

	// Temp code for finding max breath value on a device; may vary from device to device?
//	static float max_breath_value = 0.00001;
//	if(max_breath_value < next_breath_value)
//	{
//		max_breath_value = next_breath_value;
//	}
//	assert (max_breath_value > 0);
//	next_breath_value /= max_breath_value;
//	next_breath_value *= 1.5;

	// This scaling value was found experimentally using an ipod touch v4.
	next_breath_value /= .002;

	// Cap value at 1.0
	if (next_breath_value > 1.0)
	{
		next_breath_value = 1.0;
	}

	// Change breath pressure
	if (next_breath_value > 0.0)
	{
		instrument->startBlowing( next_breath_value , 1.0 );
	}
	else
	{
		instrument->stopBlowing(1.0);
	}

	// Slew values once per frame
	SaxMax::slew_values();	
}

void SaxMax::slew_values()
{
	// Slew parameters for instrument
	for (int i=0; i<NUM_PARAMETERS; ++i)
	{
		if (i != BREATH_PRESSURE)
		{
			control_parameters[i]->slew();
			instrument->controlChange(ControlParameter::get_parameter_from_enum(i),control_parameters[i]->value);
		}
	}
}

void SaxMax::set_dims(GLint width, GLint height)
{
	g_gfxWidth = width;
    g_gfxHeight = height;
}

stk::Saxofony* SaxMax::get_instrument()
{
	return instrument;
}

void SaxMax::render()
{
	// Reset projection matrix
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrthof(-1.0, 1.0, -1.5, 1.5, -1.0, 1.0);
	
	// Model view
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	// Set clear color
	glClearColor(control_parameters[BLOW_POSITION]->value/(2*128), 0.5 - control_parameters[BLOW_POSITION]->value/(2*128),0,1.0);
	glClear(GL_COLOR_BUFFER_BIT);
	
	// Map the viewport
	MoGfx::ortho(g_gfxWidth, g_gfxHeight, 1);
	glEnableClientState(GL_VERTEX_ARRAY);

	draw_waveform();
	draw_toneholes();
}

bool SaxMax::init()
{
	bool result = true;
	stk::Stk::setSampleRate(SAMPLE_RATE);
	
	control_parameters.clear();
	for (int i=0; i<NUM_PARAMETERS; ++i)
	{
		control_parameters.push_back(new ControlParameter());
		control_parameters[i]->value = control_parameters[i]->target = 0.5;
		control_parameters[i]->slew_rate = 0.1;
	}

	// Initialize instrument and parameters
	instrument = new stk::Saxofony(100);

	// Filters for 
	mic_filter = new stk::OnePole(0.997);
	mic_filter2 = new stk::OnePole(0.9);
	rev = new stk::NRev();

	// Set up control parameters
	reset_control_parameters();
	
	// Run NoteOn and NoteOff to set amplitudes:
	instrument->noteOn(0.0, 1.0-0.001);
	instrument->noteOff(1.0);
	
	MoAccel::addCallback( accelCallback, NULL);
	MoTouch::addCallback( touchCallback, NULL);
	if (result)
	{
		result = MoAudio::init(SAMPLE_RATE, FRAMESIZE, NUM_CHANNELS);
	}
	if (result)
	{
		result = MoAudio::start(audio_callback, NULL);
	}

	// Enable blending to allow transparent textures to overlap correctly
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

	// Load texture
	glGenTextures(1, texture);
	glBindTexture(GL_TEXTURE_2D, texture[0]);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	MoGfx::loadTexture(@"texture1", @"png");		
	
	return result;
}

void SaxMax::close()
{
	MoAudio::shutdown();
	for (int i=0; i<control_parameters.size(); ++i)
	{
		delete control_parameters[i];
	}
	delete mic_filter;
	delete instrument;
}

float SaxMax::midi_to_frequency(int note)
{
    assert (note >=0);
    assert (note <=127);
	
    // Reference frequency is A440, note number 69
    return (440.0 * pow(2,((float)note-69.0)/12.0));
}

void SaxMax::calculate_frequency()
{
	int note = 0;
	bool update = true;

	// Pack into an int
	int status = 0;
	if (tonehole_status[1])
	{
		status += 8;
	}
	if (tonehole_status[2])
	{
		status += 4;
	}
	if (tonehole_status[3])
	{
		status += 2;
	}
	if (tonehole_status[4])
	{
		status += 1;
	}
	
	// Switch on value
	switch(status)
	{
		case 0xf:
			note = 64;
			break;
		case 0xe:
			note = 65;
			break;
		case 0xd:
			note = 66;
			break;
		case 0xc:
			note = 67;
			break;
		case 0xa:
			note = 68;
			break;
		case 0x8:
			note = 69;
			break;
		case 0xb:
			note = 70;
			break;
		case 0x9:
			note = 71;
			break;
		case 0x3:
			note = 72;
			break;
		case 0x2:
			note = 73;
			break;
		case 0x6:
			note = 74;
			break;
		case 0x7:
			note = 75;
			break;
		default:
			update = false;
			break;
	};
	
	// Use last tonehole as an octave switch
	if (tonehole_status[0])
	{
		note -= 12;
	}

	if (update)
	{
		current_frequency = midi_to_frequency(note);
		instrument->setFrequency(current_frequency);
	}
}

void SaxMax::reset_control_parameters()
{
	control_parameters[REED_STIFFNESS]->value = control_parameters[REED_STIFFNESS]->target = 0.3 * 128.0;
	control_parameters[REED_APETURE]->value = control_parameters[REED_APETURE]->target = 0.7 * 128.0;
	control_parameters[VIBRATO_GAIN]->value = control_parameters[VIBRATO_GAIN]->target = 0.1 * 128.0;
	control_parameters[VIBRATO_FREQUENCY]->value = control_parameters[VIBRATO_FREQUENCY]->target = 0.3 * 128.0;
	control_parameters[NOISE_GAIN]->value = control_parameters[NOISE_GAIN]->target = 0.2 * 128.0;
	control_parameters[BREATH_PRESSURE]->value = control_parameters[BREATH_PRESSURE]->target = 0.8 * 128.0;
	control_parameters[BLOW_POSITION]->value = control_parameters[BLOW_POSITION]->target = 0.2 * 128.0;
	reverb_t60 = 1.0;
	set_reverb(reverb_t60);
}

void SaxMax::set_control(int parameter, float value)
{
	control_parameters[parameter]->target = value;
}

float SaxMax::get_control(int parameter)
{
	return control_parameters[parameter]->target;
}

void SaxMax::set_reverb(float value)
{
	assert (value > 0);
	reverb_t60 = value;
	rev->setT60(value);
}

float SaxMax::get_reverb()
{
	return reverb_t60;
}

void SaxMax::draw_waveform()
{
	// Cycle colors.
	static float color_angles[3] = {0.0, 0.0, 0.0};
	color_angles[0] += .005;
	color_angles[1] += .007;
	color_angles[2] += .009;
	for (int i=0; i<3; i++)
	{
		while(color_angles[i] > 3.14159)
		{
			color_angles[i] -= 3.14159;
		}
	}
	
	float xinc = (g_gfxHeight) / (float)NUM_VERTICES;
	
	// Draw the currently playing waveform
	for (int i=0; i < NUM_VERTICES; ++i)
	{
		m_line_scratch[i*2] = i*xinc;
		m_line_scratch[i*2+1] = (g_gfxWidth / 2.0) + (((g_gfxWidth-1)/2.0) * m_vertices[i]);
	}

	// Draw line
	glPushMatrix();
	glVertexPointer(2, GL_FLOAT, 0, m_line_scratch);
	glEnableClientState( GL_VERTEX_ARRAY );
	glColor4f( sin(color_angles[0]), sin(color_angles[1]), sin(color_angles[2]), 1.0);
	glLineWidth(2.0);
	glDrawArrays( GL_LINE_STRIP, 0, NUM_VERTICES );
	glPopMatrix();
}

void SaxMax::draw_toneholes()
{
	static const GLfloat region1[] =
	{
		0.0, 0.0,
		g_gfxHeight/3.0, 0.0,
		0.0, g_gfxWidth/2.0,
		g_gfxHeight/3.0, g_gfxWidth/2.0,
	};

	static const GLfloat region2[] =
	{
		0.0, g_gfxWidth/2.0,
		g_gfxHeight/3.0, g_gfxWidth/2.0,
		0.0, g_gfxWidth,
		g_gfxHeight/3.0, g_gfxWidth,
	};
	
	static const GLfloat region3[] =
	{
		g_gfxHeight/3.0, 0.0,
		g_gfxHeight*2/3.0, 0.0,
		g_gfxHeight/3.0, g_gfxWidth/2.0,
		g_gfxHeight*2/3.0, g_gfxWidth/2.0,
	};
	
	static const GLfloat region4[] =
	{
		g_gfxHeight/3.0, g_gfxWidth/2.0,
		g_gfxHeight*2/3.0, g_gfxWidth/2.0,
		g_gfxHeight/3.0, g_gfxWidth,
		g_gfxHeight*2/3.0, g_gfxWidth,
	};
	
	static const GLfloat region5[] =
	{
		g_gfxHeight*2/3.0, 0.0,
		g_gfxHeight, 0.0,
		g_gfxHeight*2/3.0, g_gfxWidth,
		g_gfxHeight, g_gfxWidth,
	};
	
	
	
	static const GLfloat texCoords[] =
	{
		0,1,
		1,1,
		0,0,
		1,0
	};

	
	static float angle = 0.0;
	angle += 0.05;
	while(angle > 3.14159)
	{
		angle -= 3.14159;
	}
	
	
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, texture[0]);
	glTexCoordPointer(2, GL_FLOAT, 0, texCoords);
	glEnableClientState(GL_TEXTURE_COORD_ARRAY);
	glPushMatrix();

	glColor4f(1.0, 1.0, 0.2, 0.35 + 0.5*sin(angle));

	glVertexPointer( 2, GL_FLOAT, 0, region1 );
	glDrawArrays( GL_TRIANGLE_STRIP, 0, 4 );

	glVertexPointer( 2, GL_FLOAT, 0, region2 );
	glDrawArrays( GL_TRIANGLE_STRIP, 0, 4 );

	glVertexPointer( 2, GL_FLOAT, 0, region3 );
	glDrawArrays( GL_TRIANGLE_STRIP, 0, 4 );

	glVertexPointer( 2, GL_FLOAT, 0, region4 );
	glDrawArrays( GL_TRIANGLE_STRIP, 0, 4 );

	glVertexPointer( 2, GL_FLOAT, 0, region5 );
	glDrawArrays( GL_TRIANGLE_STRIP, 0, 4 );
	
	glPopMatrix();
	glDisable(GL_TEXTURE_2D);
}
